@@ -2,7 +2,7 @@ import inspect
 import sys
 import textwrap
 import typing
-from typing import get_type_hints, TypeVar, Any, AnyStr, Tuple
+from typing import Any, AnyStr, Tuple, TypeVar, get_type_hints
 
 from sphinx.util import logging
 from sphinx.util.inspect import signature as Signature
@@ -72,17 +72,6 @@ def get_annotation_args(annotation, module: str, class_name: str) -> Tuple:
     # Special cases
     if class_name in ('Pattern', 'Match') and hasattr(annotation, 'type_var'):  # Python < 3.7
         return annotation.type_var,
-    elif class_name == 'Callable' and hasattr(annotation, '__result__'):  # Python < 3.5.3
-        argtypes = (Ellipsis,) if annotation.__args__ is Ellipsis else annotation.__args__
-        return argtypes + (annotation.__result__,)
-    elif class_name == 'Union' and hasattr(annotation, '__union_params__'):  # Union on Python 3.5
-        return annotation.__union_params__
-    elif class_name == 'Tuple' and hasattr(annotation, '__tuple_params__'):  # Tuple on Python 3.5
-        params = annotation.__tuple_params__
-        if getattr(annotation, '__tuple_use_ellipsis__', False):
-            params += (Ellipsis,)
-
-        return params
     elif class_name == 'ClassVar' and hasattr(annotation, '__type__'):  # ClassVar on Python < 3.7
         return annotation.__type__,
     elif class_name == 'NewType' and hasattr(annotation, '__supertype__'):
@@ -95,10 +84,12 @@ def get_annotation_args(annotation, module: str, class_name: str) -> Tuple:
     return getattr(annotation, '__args__', ())
 
 
-def format_annotation(annotation, fully_qualified: bool = False) -> str:
+def format_annotation(annotation,
+                      fully_qualified: bool = False,
+                      simplify_optional_unions: bool = True) -> str:
     # Special cases
     if annotation is None or annotation is type(None):  # noqa: E721
-        return '``None``'
+        return ':py:obj:`None`'
     elif annotation is Ellipsis:
         return '...'
 
@@ -114,7 +105,7 @@ def format_annotation(annotation, fully_qualified: bool = False) -> str:
         class_name = get_annotation_class_name(annotation, module)
         args = get_annotation_args(annotation, module, class_name)
     except ValueError:
-        return str(annotation)
+        return str(annotation).strip("'")
 
     # Redirect all typing_extensions types to the stdlib typing module
     if module == 'typing_extensions':
@@ -130,23 +121,72 @@ def format_annotation(annotation, fully_qualified: bool = False) -> str:
     if full_name == 'typing.NewType':
         args_format = '\\(:py:data:`~{name}`, {{}})'.format(name=annotation.__name__)
         role = 'func'
-    elif full_name == 'typing.Union' and len(args) == 2 and type(None) in args:
-        full_name = 'typing.Optional'
-        args = tuple(x for x in args if x is not type(None))  # noqa: E721
+    elif full_name == 'typing.Union' and type(None) in args:
+        if len(args) == 2:
+            full_name = 'typing.Optional'
+            args = tuple(x for x in args if x is not type(None))  # noqa: E721
+        elif not simplify_optional_unions:
+            full_name = 'typing.Optional'
+            args_format = '\\[:py:data:`{prefix}typing.Union`\\[{{}}]]'.format(prefix=prefix)
+            args = tuple(x for x in args if x is not type(None))  # noqa: E721
     elif full_name == 'typing.Callable' and args and args[0] is not ...:
-        formatted_args = '\\[\\[' + ', '.join(format_annotation(arg) for arg in args[:-1]) + ']'
-        formatted_args += ', ' + format_annotation(args[-1]) + ']'
+        formatted_args = '\\[\\[' + ', '.join(
+            format_annotation(
+                arg, simplify_optional_unions=simplify_optional_unions)
+            for arg in args[:-1]) + ']'
+        formatted_args += ', ' + format_annotation(
+            args[-1], simplify_optional_unions=simplify_optional_unions) + ']'
     elif full_name == 'typing.Literal':
         formatted_args = '\\[' + ', '.join(repr(arg) for arg in args) + ']'
 
     if args and not formatted_args:
-        formatted_args = args_format.format(', '.join(format_annotation(arg, fully_qualified)
-                                                      for arg in args))
+        formatted_args = args_format.format(', '.join(
+            format_annotation(arg, fully_qualified, simplify_optional_unions)
+            for arg in args))
 
     return ':py:{role}:`{prefix}{full_name}`{formatted_args}'.format(
         role=role, prefix=prefix, full_name=full_name, formatted_args=formatted_args)
 
 
+# reference: https://github.com/pytorch/pytorch/pull/46548/files
+def normalize_source_lines(sourcelines: str) -> str:
+    """
+    This helper function accepts a list of source lines. It finds the
+    indentation level of the function definition (`def`), then it indents
+    all lines in the function body to a point at or greater than that
+    level. This allows for comments and continued string literals that
+    are at a lower indentation than the rest of the code.
+    Arguments:
+        sourcelines: source code
+    Returns:
+        source lines that have been correctly aligned
+    """
+    sourcelines = sourcelines.split("\n")
+
+    def remove_prefix(text, prefix):
+        return text[text.startswith(prefix) and len(prefix):]
+
+    # Find the line and line number containing the function definition
+    for i, l in enumerate(sourcelines):
+        if l.lstrip().startswith("def"):
+            idx = i
+            break
+    else:
+        return "\n".join(sourcelines)
+    fn_def = sourcelines[idx]
+
+    # Get a string representing the amount of leading whitespace
+    whitespace = fn_def.split("def")[0]
+
+    # Add this leading whitespace to all lines before and after the `def`
+    aligned_prefix = [whitespace + remove_prefix(s, whitespace) for s in sourcelines[:idx]]
+    aligned_suffix = [whitespace + remove_prefix(s, whitespace) for s in sourcelines[idx + 1:]]
+
+    # Put it together again
+    aligned_prefix.append(fn_def)
+    return "\n".join(aligned_prefix + aligned_suffix)
+
+
 def process_signature(app, what: str, name: str, obj, options, signature, return_annotation):
     if not callable(obj):
         return
@@ -165,8 +205,20 @@ def process_signature(app, what: str, name: str, obj, options, signature, return
         for param in signature.parameters.values()
     ]
 
-    # The generated dataclass __init__() is weird and needs the second condition
-    if '<locals>' in obj.__qualname__ and not (what == 'method' and name.endswith('.__init__')):
+    # The generated dataclass __init__() and class are weird and need extra checks
+    # This helper function operates on the generated class and methods
+    # of a dataclass, not an instantiated dataclass object. As such,
+    # it cannot be replaced by a call to `dataclasses.is_dataclass()`.
+    def _is_dataclass(name: str, what: str, qualname: str) -> bool:
+        if what == 'method' and name.endswith('.__init__'):
+            # generated __init__()
+            return True
+        if what == 'class' and qualname.endswith('.__init__'):
+            # generated class
+            return True
+        return False
+
+    if '<locals>' in obj.__qualname__ and not _is_dataclass(name, what, obj.__qualname__):
         logger.warning(
             'Cannot treat a function defined as a local function: "%s"  (use @functools.wraps)',
             name)
@@ -257,7 +309,8 @@ def backfill_type_hints(obj, name):
         return children[0]
 
     try:
-        obj_ast = ast.parse(textwrap.dedent(inspect.getsource(obj)), **parse_kwargs)
+        obj_ast = ast.parse(textwrap.dedent(
+            normalize_source_lines(inspect.getsource(obj))), **parse_kwargs)
     except (OSError, TypeError):
         return {}
 
@@ -370,7 +423,9 @@ def process_docstring(app, what, name, obj, options, lines):
                 argname = '{}\\_'.format(argname[:-1])
 
             formatted_annotation = format_annotation(
-                annotation, fully_qualified=app.config.typehints_fully_qualified)
+                annotation,
+                fully_qualified=app.config.typehints_fully_qualified,
+                simplify_optional_unions=app.config.simplify_optional_unions)
 
             searchfor = [':{} {}:'.format(field, argname)
                          for field in ('param', 'parameter', 'arg', 'argument')]
@@ -397,7 +452,9 @@ def process_docstring(app, what, name, obj, options, lines):
                 return
 
             formatted_annotation = format_annotation(
-                type_hints['return'], fully_qualified=app.config.typehints_fully_qualified)
+                type_hints['return'], fully_qualified=app.config.typehints_fully_qualified,
+                simplify_optional_unions=app.config.simplify_optional_unions
+            )
 
             insert_index = len(lines)
             for i, line in enumerate(lines):
@@ -427,6 +484,7 @@ def setup(app):
     app.add_config_value('always_document_param_types', False, 'html')
     app.add_config_value('typehints_fully_qualified', False, 'env')
     app.add_config_value('typehints_document_rtype', True, 'env')
+    app.add_config_value('simplify_optional_unions', True, 'env')
     app.connect('builder-inited', builder_ready)
     app.connect('autodoc-process-signature', process_signature)
     app.connect('autodoc-process-docstring', process_docstring)
