@@ -4,20 +4,15 @@ import sys
 import textwrap
 import typing
 from typing import (
-    Any, AnyStr, Callable, Dict, Generic, Mapping, NewType, Optional, Pattern, Match, Tuple,
-    TypeVar, Union, Type)
+    IO, Any, AnyStr, Callable, Dict, Generic, Mapping, Match, NewType, Optional, Pattern, Tuple,
+    Type, TypeVar, Union)
 
 import pytest
 import typing_extensions
 
 from sphinx_autodoc_typehints import (
-    format_annotation, process_docstring, get_annotation_module, get_annotation_class_name,
-    get_annotation_args)
-
-try:
-    from typing import IO
-except ImportError:
-    from typing.io import IO
+    format_annotation, get_annotation_args, get_annotation_class_name, get_annotation_module,
+    process_docstring)
 
 T = TypeVar('T')
 U = TypeVar('U', covariant=True)
@@ -126,11 +121,15 @@ def test_parse_annotation(annotation, module, class_name, args):
     (Union,                         ':py:data:`~typing.Union`'),
     (Union[str, bool],              ':py:data:`~typing.Union`\\[:py:class:`str`, '
                                     ':py:class:`bool`]'),
+    (Union[str, bool, None],        ':py:data:`~typing.Union`\\[:py:class:`str`, '
+                                    ':py:class:`bool`, :py:obj:`None`]'),
     pytest.param(Union[str, Any],   ':py:data:`~typing.Union`\\[:py:class:`str`, '
                                     ':py:data:`~typing.Any`]',
                  marks=pytest.mark.skipif((3, 5, 0) <= sys.version_info[:3] <= (3, 5, 2),
                                           reason='Union erases the str on 3.5.0 -> 3.5.2')),
     (Optional[str],                 ':py:data:`~typing.Optional`\\[:py:class:`str`]'),
+    (Optional[Union[str, bool]],    ':py:data:`~typing.Union`\\[:py:class:`str`, '
+                                    ':py:class:`bool`, :py:obj:`None`]'),
     (Callable,                      ':py:data:`~typing.Callable`'),
     (Callable[..., int],            ':py:data:`~typing.Callable`\\[..., :py:class:`int`]'),
     (Callable[[int], int],          ':py:data:`~typing.Callable`\\[\\[:py:class:`int`], '
@@ -158,6 +157,27 @@ def test_format_annotation(inv, annotation, expected_result):
     result = format_annotation(annotation)
     assert result == expected_result
 
+    # Test with the "simplify_optional_unions" flag turned off:
+    if re.match(r'^:py:data:`~typing\.Union`\\\[.*``None``.*\]', expected_result):
+        # strip None - argument and copy string to avoid conflicts with
+        # subsequent tests
+        expected_result_not_simplified = expected_result.replace(', ``None``', '')
+        # encapsulate Union in typing.Optional
+        expected_result_not_simplified = ':py:data:`~typing.Optional`\\[' + \
+            expected_result_not_simplified
+        expected_result_not_simplified += ']'
+        assert format_annotation(annotation, simplify_optional_unions=False) == \
+            expected_result_not_simplified
+
+        # Test with the "fully_qualified" flag turned on
+        if 'typing' in expected_result_not_simplified:
+            expected_result_not_simplified = expected_result_not_simplified.replace('~typing',
+                                                                                    'typing')
+            assert format_annotation(annotation,
+                                     fully_qualified=True,
+                                     simplify_optional_unions=False) == \
+                expected_result_not_simplified
+
     # Test with the "fully_qualified" flag turned on
     if 'typing' in expected_result or __name__ in expected_result:
         expected_result = expected_result.replace('~typing', 'typing')
@@ -223,10 +243,14 @@ def test_sphinx_output(app, status, warning, always_document_param_types):
     assert 'Cannot resolve forward reference in type annotations of ' in warnings, warnings
 
     format_args = {}
-    if always_document_param_types:
-        format_args['undoc_params'] = '\n\n   Parameters:\n      **x** ("int") --'
-    else:
-        format_args['undoc_params'] = ''
+    for indentation_level in range(2):
+        key = f'undoc_params_{indentation_level}'
+        if always_document_param_types:
+            format_args[key] = textwrap.indent(
+                '\n\n   Parameters:\n      **x** ("int") --', '   ' * indentation_level
+            )
+        else:
+            format_args[key] = ''
 
     text_path = pathlib.Path(app.srcdir) / '_build' / 'text' / 'index.txt'
     with text_path.open('r') as f:
@@ -423,6 +447,10 @@ def test_sphinx_output(app, status, warning, always_document_param_types):
               Return type:
                  "int"
 
+           method_without_typehint(x)
+
+              Method docstring.
+
         dummy_module.function_with_typehint_comment_not_inline(x=None, *y, z, **kwargs)
 
            Function docstring.
@@ -469,11 +497,19 @@ def test_sphinx_output(app, status, warning, always_document_param_types):
 
         dummy_module.undocumented_function(x)
 
-           Hi{undoc_params}
+           Hi{undoc_params_0}
 
            Return type:
               "str"
 
+        class dummy_module.DataClass(x)
+
+           Class docstring.{undoc_params_0}
+
+           __init__(x)
+
+              Initialize self.  See help(type(self)) for accurate signature.{undoc_params_1}
+
         @dummy_module.Decorator(func)
 
            Initializer docstring.
